print("import: 'com'")
import com

print("import: 'sparknlp'")
import sparknlp

print("import: 'sparknlp.annotator'")
import sparknlp.annotator

print("import: 'sparknlp.base'")
import sparknlp.base

print("import: 'sparknlp.common'")
import sparknlp.common

print("import: 'sparknlp.internal'")
import sparknlp.internal

print("import: 'sparknlp.logging'")
import sparknlp.logging

print("import: 'sparknlp.pretrained'")
import sparknlp.pretrained

print("import: 'sparknlp.training'")
import sparknlp.training

